(* Theorems about CPS conversion *)

Require Import Coq.Lists.List.
Require Import ListFacts.
Require Import "Variable".
Require Import Tactics.
Require Import SourceLang.
Require Import CpsLang.
Require Import Cps.

Open Scope ctype_scope.


Lemma cps_ctxt_preserves_in :
  forall (C : sctxt) (x : var) (T : stype),
  in' x T C ->
  in' x (cps_type T) (cps_ctxt C).
Proof.
  induction C as [ | [ x' T' ] C' IH ];
  intros x T H.
    case H.

    destruct H as [ [ <- <- ] | [ H1 H2 ] ]; fold (@in' var stype) in *.
      left; split; reflexivity.

      right; split.
        exact H1.

        apply IH, H2.
Qed.


Lemma open_c_cps_term_id :
  forall (n : nat) (t : sterm) (t' : cterm),
  open_cterm_c_n n t' (cps_term t) = cps_term t.
Proof.
  intros n t; revert n;
  induction t as [ n' | x | | t1 IH1 | t1 IH1 t2 IH2 ];
  intros n t';
  simpl;
  repeat f_equal;
  auto.
Qed.


Lemma open_sfvar_cps_term_swap :
  forall (x : var) (t : sterm),
  open_cterm_s (ct_sfvar x) (cps_term t) = cps_term (open_sterm (st_fvar x) t).
Proof.
  intros x t;
  unfold open_sterm, open_cterm_s.
  generalize 0.
  induction t as [ n' | x' | | t1 IH1 | t1 IH1 t2 IH2 ];
  intro n;
  simpl;
  [ destruct (Peano_dec.eq_nat_dec n n') | | | | ];
  repeat f_equal;
  auto.
Qed.


Theorem cps_term_preserves_typing :
  forall (Cs : sctxt) (T : stype) (t : sterm),
  sterm_typing Cs t T ->
  forall Cc : cctxt,
  cterm_typing (cps_ctxt Cs) Cc (cps_term t) (! ! cps_type T).
Proof.
  induction 1 as [ Cs x T H1 H2
                 | Cs
                 | Cs l TA TR t H1 IH1
                 | Cs TA TR tf ta H1 IH1 H2 IH2 ];
  intro Cc;
  simpl.

  (* case stt_fvar *)
  apply ctt_cabs with (l := domain_of_cctxt Cc). intros x1 H3.
  unfold open_cterm_c; simpl open_cterm_c_n.
  apply ctt_app with (TA := cps_type T).
    apply ctt_cfvar;
    auto_in constructor.

    apply ctt_sfvar.
    apply cps_ctxt_preserves_in, H1.

  (* case stt_unit *)
  apply ctt_cabs with (l := domain_of_cctxt Cc). intros x H1.
  unfold open_cterm_c; simpl open_cterm_c_n.
  apply ctt_app with (TA := cps_type sy_unit).
    apply ctt_cfvar;
    auto_in constructor.

    apply ctt_unit.

  (* case stt_abs *)
  apply ctt_cabs with (l := domain_of_cctxt Cc). intros x1 H2.
  unfold open_cterm_c; simpl open_cterm_c_n.
  apply ctt_app with (TA := cps_type (sy_func TA TR)).
    apply ctt_cfvar;
    auto_in constructor.

    apply ctt_sabs with (l := l); fold cps_type. intros x2 H3.
    rewrite open_c_cps_term_id.
    rewrite open_sfvar_cps_term_swap.
    apply (IH1 x2 H3).

  (* case stt_app *)
  apply ctt_cabs with (l := domain_of_cctxt Cc). intros x1 H3.
  unfold open_cterm_c; simpl.
  rewrite !open_c_cps_term_id.
  apply ctt_app with (TA := ! cps_type (sy_func TA TR)).
    apply IH1.

    apply ctt_cabs with (l := x1 :: domain_of_cctxt Cc). intros x2 H4.
    unfold open_cterm_c; simpl open_cterm_c_n.
    rewrite open_c_cps_term_id.
    apply ctt_app with (TA := ! cps_type TA).
      apply IH2.

      apply ctt_cabs with (l := x2 :: x1 :: domain_of_cctxt Cc). intros x3 H5.
      unfold open_cterm_c; simpl open_cterm_c_n.
      apply ctt_app with (TA := ! cps_type TR);
      [ apply ctt_app with (TA := cps_type TA) | ];
      apply ctt_cfvar; simpl; auto_in.
Qed.
