(* Definition of the source language *)

Require Import Coq.Arith.Peano_dec.
Require Import Coq.Lists.List.
Require Import "Variable".
Require Import ListFacts.


Section Terms.

(* term *)
Inductive sterm : Set :=
  | st_bvar (n : nat)  (* bound variable *)
  | st_fvar (x : var)  (* free variable *)
  | st_unit
  | st_abs (t : sterm)
  | st_app (tf ta : sterm).

(* substitute bound variable n in term t with term t' *)
Fixpoint open_sterm_n (n : nat) (t' t : sterm) : sterm :=
  match t with
  | st_bvar n1 =>
      if eq_nat_dec n n1 then t' else st_bvar n1
  | st_fvar x =>
      st_fvar x
  | st_unit =>
      st_unit
  | st_abs t1 =>
      st_abs (open_sterm_n (S n) t' t1)
  | st_app t1 t2 =>
      st_app (open_sterm_n n t' t1) (open_sterm_n n t' t2)
  end.
Definition open_sterm : sterm -> sterm -> sterm :=
  open_sterm_n O.

End Terms.


Section Types.

(* type *)
Inductive stype : Set :=
  | sy_unit
  | sy_func (TA TR : stype).

End Types.


Section Contexts.

(* typing context *)
Definition sctxt : Set :=
  list (var * stype).

Definition domain_of_sctxt : sctxt -> list var :=
  map (@fst _ _).

End Contexts.


Section Rules.

Inductive sterm_typing : sctxt -> sterm -> stype -> Prop :=
  | stt_fvar :
      forall (C : sctxt) (x : var) (T : stype),
      in' x T C ->
      sterm_typing C (st_fvar x) T
  | stt_unit :
      forall C : sctxt,
      sterm_typing C st_unit sy_unit
  | stt_abs :
      forall (C : sctxt) (l : list var) (TA TR : stype) (t : sterm),
      (forall x : var, ~ In x l ->
       sterm_typing ((x, TA) :: C) (open_sterm (st_fvar x) t) TR) ->
      sterm_typing C (st_abs t) (sy_func TA TR)
  | stt_app :
      forall (C : sctxt) (TA TR : stype) (tf ta : sterm),
      sterm_typing C tf (sy_func TA TR) ->
      sterm_typing C ta TA ->
      sterm_typing C (st_app tf ta) TR.

End Rules.
