(* Auxiliary tactics *)

Require Import Coq.Arith.Peano_dec.
Require Import Coq.Lists.List.
Require Import ListFacts.
Require Import "Variable".


Ltac destruct_eq_dec dec t :=
  let H := fresh "H" in
  case (dec t t); [ intro H; clear H | clear; congruence ].
Ltac destruct_eq_dec_all :=
  repeat
    match goal with
    | H : context[eq_nat_dec ?x ?x] |- _ =>
        destruct_eq_dec eq_nat_dec x
    | |- context[eq_nat_dec ?x ?x] =>
        destruct_eq_dec eq_nat_dec x
    | H : context[eq_var_dec ?x ?x] |- _ =>
        destruct_eq_dec eq_var_dec x
    | |- context[eq_var_dec ?x ?x] =>
        destruct_eq_dec eq_var_dec x
    end.

Ltac auto_logic_simpl :=
  match goal with
  | |- True =>
      exact I
  | H : False |- _ =>
      case H
  | |- forall _ : _, _ =>
      intro
  | H : _ /\ _ |- _ =>
      destruct H
  | |- _ /\ _ =>
      split
  | |- context[not] =>
      unfold not in |- *
  | H : context[not] |- _ =>
      unfold not in H |-
  | |- context[iff] =>
      unfold iff in |- *
  | H : context[iff] |- _ =>
      unfold iff in H |-
  end.
Ltac auto_logic tac :=
  match goal with
  | _ =>
      progress tac; auto_logic tac
  | _ =>
      progress auto_logic_simpl; auto_logic tac
  | H : _ |- _ =>
      apply H; clear H; auto_logic tac
  | H : _ \/ _ |- _ =>
      destruct H; auto_logic tac
  | |- _ \/ _ =>
      left; auto_logic tac
  | |- _ \/ _ =>
      right; auto_logic tac
  end.

Ltac auto_in_simpl :=
  match goal with
  | H : In ?x nil |- _ =>
      case H
  | |- In ?x ?l =>
      apply in_eq
  | |- In ?x (?y :: ?l) =>
      simpl In in |- *
  | H : In ?x (?y :: ?l) |- _ =>
      apply in_inv in H
  | |- In ?x (?l1 ++ ?l2) =>
      apply in_or_app
  | H : In ?x (?l1 ++ ?l2) |- _ =>
      apply in_app_or in H
  | |- context[incl] =>
      unfold incl in |- *
  | H : context[incl] |- _ =>
      unfold incl in H |-
  | |- In ?x (remove ?eq_dec ?x' ?l) =>
      apply neq_remove_In
  | H : In ?x (remove ?eq_dec ?x' ?l) |- _ =>
      apply remove_remove_In in H
  end.
Tactic Notation "auto_in" tactic(tac) :=
  auto_logic ltac:(auto_in_simpl || tac).
Tactic Notation "auto_in" :=
  auto_in auto.
